from .download import download_file
from pathlib import Path
down_path = Path(__file__).parent
xz_url = "addons_file"+"/"+down_path.name+"/"
download_file(xz_url + "data.blend",down_path)

import math
import os

import bpy
from bpy.app.handlers import persistent
from bpy.types import Menu, Operator, Panel, UIList

bl_info = {
    "name": "Post FX",
    "author": "Thatimster",
    "version": (2, 1),
    "blender": (4, 0, 2),
    "location": "Compositor Toolbar",
    "description": "Next Level Compositing!",
    "warning": "",
    "wiki_url": "",
    "tracker_url": "https://www.thatimster.com/contact.html",
    "category": "Compositor"
}

func_map = {

    "rl": ("Render Layer", "RENDERLAYERS", "CompositorNodeRLayers", "red"),
    "im": ("Image", "OUTLINER_OB_IMAGE", "CompositorNodeImage", "red"),
    "no": ("Noise", "TEXTURE", "CompositorNodeGroup", "red"),
    "co": ("Coordinates", "ORIENTATION_VIEW", "CompositorNodeGroup", "red"),
    "ic": ("Control", "IMAGE_PLANE", "CompositorNodeGroup", "orange"),

    "op": ("Composite", "OUTPUT", "COMPOSITE", "CompositorNodeComposite", "burgundy"),
    "pr": ("Preview", "IMAGE_BACKGROUND", "VIEWER", "CompositorNodeViewer", "burgundy"),
    "ba": ("Before / After", "SEQ_SPLITVIEW", "SPLITVIEWER", "CompositorNodeSplitViewer", "burgundy"),

    "br": ("Brightness", "OUTLINER_OB_LIGHT", "green"),
    "cl": ("Clarity", "MOD_THICKNESS", "green"),
    "cn": ("Contrast", "SHADING_RENDERED", "green"),
    "sh": ("Shadows", "LIGHT", "green"),
    "sp": ("Sharpness", "SHARPCURVE", "green"),
    "sa": ("Hue/Sat/Lum", "MOD_HUE_SATURATION", "green"),
    "ti": ("Tint", "MATFLUID", "green"),
    "vi": ("Vibrance", "MOD_ARRAY", "green"),
    "wa": ("Warmth", "FREEZE", "green"),
    "di": ("Diffusion", "IMAGE_RGB_ALPHA", "green"),
    "fa": ("Fade", "IMAGE_RGB", "green"),
    "gr": ("Graduate", "NODE_TEXTURE", "green"),
    "le": ("Levels", "SEQ_HISTOGRAM", "green"),
    "se": ("Selective HSL", "NLA", "green"),
    "ca": ("Calibration", "MODIFIER", "green"),

    "f1": ("Orange + Teal", "COLORSET_02_VEC", "blue"),
    "f2": ("Punchy BW", "COLORSET_10_VEC", "blue"),
    "f3": ("Dreamy", "COLORSET_05_VEC", "blue"),
    "f4": ("Gritty", "COLORSET_08_VEC", "blue"),
    "f5": ("Matrix", "COLORSET_07_VEC", "blue"),
    "f6": ("VCR", "COLORSET_06_VEC", "blue"),
    "f7": ("Detail Boost", "COLORSET_03_VEC", "blue"),
    "f8": ("Sepia", "COLORSET_14_VEC", "blue"),
    "f9": ("Colorize", "COLORSET_04_VEC", "blue"),
    "f10": ("CCTV", "COLORSET_13_VEC", "blue"),

    "ln": ("Lens", "VIEW_CAMERA", "purple"),
    "fr": ("Fringing", "GP_MULTIFRAME_EDITING", "purple"),
    "vg": ("Vignette", "IMAGE_PLANE", "purple"),
    "ts": ("Tilt Shift", "ALIGN_MIDDLE", "purple"),
    "rb": ("Radial Blur", "PROP_CON", "purple"),
    "de": ("Double Exposure", "IMAGE_ZDEPTH", "purple"),

    "fl": ("Flare Builder", "LIGHT_SUN", "purple"),
    "bl": ("Bloom", "OUTLINER_OB_LIGHTPROBE", "purple"),

    "ar": ("Artifacts", "MOD_DATA_TRANSFER", "purple"),
    "gl": ("Glitch", "LIBRARY_DATA_BROKEN", "purple"),
    "px": ("Pixelate", "ALIASED", "purple"),
    "ps": ("Pixel Screen", "RESTRICT_VIEW_OFF", "purple"),

    "sy": ("Symmetry", "MOD_MIRROR", "purple"),
    "tp": ("Tiny Planet", "WORLD", "purple"),

    "n": ("None", "", "")
}

color_lookup = {
    "red": (0.787, 0.286, 0.198),
    "orange": (0.8, 0.450, 0.18),
    "green": (0.129, 0.610, 0.122),
    "blue": (0.155, 0.407, 0.784),
    "purple": (0.448, 0.207, 0.688),
    "burgundy": (0.24, 0.12, 0.144)
}

enum_map = {
    "Input": ["n", "rl", "im"],
    "Adjustment": ["n", "br", "cl", "cn", "sp", "sa", "ca", "se", "ti", "vi", "wa", "di", "fa", "gr", "le", "sh"],
    "Filter": ["n", "f1", "f2", "f3", "f4", "f5", "f6", "f7", "f8", "f9"],
    "Effect": ["n", "ln", "fr", "vg", "de", "gl", "ar","ps", "ts", "rb", "sy", "tp", "px", "fl", "bl"]
}

prop_map = {
    "rl": ["scene", "layer"],
    "im": ["image"]
}

is_rendering = False
status = ""
subscribed = False
sub_owner = object()
TIME_INTERVAL = 0.3

def msgbus_update(context):
    pref = context.preferences.addons[__package__].preferences
    global status
    areas = [a for a in context.screen.areas if a.type== "IMAGE_EDITOR"]
    if len(areas) > 0 and not status and getattr(pref, "autorefresh"):
        bpy.ops.postfx.refresh()

def get_layer_types(self, context):
    scn = context.scene
    active_layer = scn.postfx_layers[scn.postfx_active_layer]
    if not active_layer:
        return
    idx = "Input"
    if scn.postfx_active_layer > 0:
        idx = getattr(active_layer, "layer_category")
    types = []
    for i, k in enumerate(enum_map[idx]):
        data = func_map[k]
        types.append((k, data[0], data[0], data[1], i))
    return types


def get_layer_categories(self, context):
    options = list(enum_map.keys())
    options.remove("Input")
    return [(k, k, k) for k in options]


def layer_selection(self, context):
    scn = context.scene
    nt = scn.node_tree
    layers = scn.postfx_layers
    if not layers:
        return
    active_layer = layers[scn.postfx_active_layer]

    # desync check
    for l in layers:
        if l.node_name and l.node_name not in nt.nodes:
            print(l.node_name, "not in scene nodes")
            l.node_name = ""
            l.icon = "ERROR"
            l.name = "None"

    if active_layer.node_name in nt.nodes:
        # bpy.ops.node.select_all(action="DESELECT")
        for n in nt.nodes:
            n.select = False
        nt.nodes.active = nt.nodes[active_layer.node_name]
        nt.nodes[active_layer.node_name].select = True

def assign_new_category(self, context):
    scn = context.scene
    layers = scn.postfx_layers
    active_layer = layers[scn.postfx_active_layer]
    active_cat = active_layer["layer_category"]
    
    bpy.ops.postfx.modifylayer(add=False)
    bpy.ops.postfx.modifylayer(add=True)
    active_layer = layers[scn.postfx_active_layer]
    active_layer["layer_category"] = active_cat

def assign_new_layer(self, context):
    scn = context.scene
    nt = scn.node_tree
    layers = scn.postfx_layers
    active_layer = layers[scn.postfx_active_layer]
    if not getattr(active_layer, "layer_type") or not getattr(active_layer, "layer_category"):
        return
    select_type = get_layer_types(self, context)[active_layer["layer_type"]]
    if select_type[0] == "n":
        return

    if not scn.use_nodes:
        scn.use_nodes = True

    # reset visibility
    for l in layers:
        l.visible = True

    pending_output = (None, None)

    # check for existing node
    if active_layer.node_name:
        target_node = nt.nodes[active_layer.node_name]
        if scn.postfx_active_layer > 0:
            # select prev node
            if len(target_node.inputs):
                prev_node = target_node.inputs["Image"].links[0].from_node
                if prev_node:
                    nt.nodes.active = prev_node
                    if target_node.outputs[0].links:
                        nt.links.new(
                            prev_node.outputs[0], target_node.outputs[0].links[0].to_node.inputs["Image"])

            nt.nodes.remove(target_node)
            active_layer.node_name = ""
        else:
            # add mechanism for changing input
            to_remove = [target_node]
            if "control" in target_node.label and target_node.inputs[0].links:
                to_remove.append(target_node.inputs[0].links[0].from_node)

            if target_node.outputs[0].links:
                pending_output = (
                    target_node.outputs[0].links[0].to_node, target_node.outputs[0].links[0].to_socket.name)

            for r in to_remove:
                nt.nodes.remove(r)
            active_layer.node_name = ""

    if scn.postfx_active_layer > 0:
        if not nt.nodes.active:
            named = [l.node_name for l in layers if l.node_name != ""]
            nt.nodes.active = nt.nodes[named[-1]]
        bpy.ops.postfx.add(choice=select_type[0], move_amount=0)
        active_layer.node_name = nt.nodes.active.name
        active_layer.name = nt.nodes.active.node_tree.name
    else:
        bpy.ops.postfx.input(choice=select_type[0], move_cursor=False)
        active_layer.node_name = nt.nodes.active.name
        active_layer.name = select_type[1]

    active_layer.icon = select_type[3]
    active_layer.type = select_type[0]

    node, socket = pending_output
    if node:
        nt.links.new(node.inputs[socket], nt.nodes.active.outputs[0])

    connect_output(self, context)


def connect_output(self, context):
    scn = context.scene
    nt = scn.node_tree
    layers = scn.postfx_layers

    pref = context.preferences.addons[__package__].preferences
    if not layers:
        return
    active_layer = layers[scn.postfx_active_layer]

    last_active_layer = [l for l in reversed(layers) if l.visible == True]

    # automatically connect output to last layer
    nt.nodes.active = nt.nodes[last_active_layer[0].node_name]
    bpy.ops.postfx.output(choice="op")
    nt.nodes.active = nt.nodes[active_layer.node_name]
    if (getattr(pref, "autorefresh")):
        bpy.ops.postfx.refresh()

def optimize_setting(context):
    scn = context.scene
    nt = scn.render
    if nt:
        # 检查并设置 edit_quality 属性
        if hasattr(nt, 'compositor_device'):
            if nt.compositor_device != "GPU" and context.space_data.type != "":
                nt.compositor_device = "GPU"
        else:
            print("xz-blender:请使用4.2版本")


def change_visibility(self, context):
    scn = context.scene
    nt = scn.node_tree
    pref = context.preferences.addons[__package__].preferences
    if self.node_name not in scn.node_tree.nodes:
        return
    socket = "Image"
    active_node = nt.nodes[self.node_name]
    if socket not in active_node.inputs:
        return
    prev_n = active_node.inputs[socket].links[0].from_socket

    if not self.visible:
        socks = [l.to_socket for l in active_node.outputs[socket].links]
        for s in socks:
            nt.links.new(prev_n.node.outputs[prev_n.name],
                         s.node.inputs[s.name])
    else:
        socks = [
            l.to_socket for l in prev_n.node.outputs[socket].links if l.to_node.name != self.node_name]
        for s in socks:
            nt.links.new(active_node.outputs[socket],
                         s.node.inputs[s.name])
        nt.links.new(
            prev_n.node.outputs[prev_n.name], active_node.inputs[socket])
    if (getattr(pref, "autorefresh")):
        bpy.ops.postfx.refresh()


def append_nodegroup(context, choice):
    sel_ng, color = (func_map[choice][0], func_map[choice][-1])
    if sel_ng not in bpy.data.node_groups:
        file_loc = os.path.join(os.path.dirname(
            __file__), "data.blend")
        with bpy.data.libraries.load(file_loc) as (data_from, data_to):
            data_to.node_groups = [sel_ng]
    current_nodes = context.space_data.edit_tree.nodes if context.space_data.type == "NODE_EDITOR" else context.scene.node_tree.nodes
    new_n = current_nodes.new(type="CompositorNodeGroup")
    new_n.node_tree = bpy.data.node_groups[sel_ng]
    new_n.use_custom_color = True
    new_n.color = color_lookup[color]
    new_n.name += "_pfx"
    new_n.label = "PostFX " + sel_ng

    return new_n


class ADD_OT_postfxoutput(Operator):
    """Adds selected output to compositor"""
    bl_idname = "postfx.output"
    bl_label = "Add Output"

    alts = ["SPLITVIEWER", "VIEWER", "COMPOSITE"]

    choice: bpy.props.StringProperty()

    @classmethod
    def poll(cls, context):
        an = context.scene.node_tree.nodes.active
        return an and (len(an.outputs) or (an.type in cls.alts and len([n for n in an.inputs if len(n.links) > 0]))) and [n for n in context.scene.node_tree.nodes if n == an]

    def find_by_type(self, nt, n_type):
        for n in nt.nodes:
            if n.type == n_type:
                return n

    def find_root_node(self, start):
        if not len(start.inputs) or not len(start.inputs[0].links):
            return start

        for l in start.inputs[0].links:
            result = self.find_root_node(l.from_node)
            if result:
                return result

    def execute(self, context):
        nt = context.scene.node_tree
        _, _, t, add_t, _ = func_map[self.choice]
        an = nt.nodes.active

        if an.type in self.alts:
            an = an.inputs[0].links[0].from_node

        rn = self.find_root_node(an)
        find = self.find_by_type(nt, t)

        an_output = an.outputs[0]
        if an_output.type != "RGBA":
            available = [o for o in an.outputs if o.type == "RGBA"]
            if available:
                an_output = available[0]

        if not find:
            # add required node
            find = nt.nodes.new(type=add_t)

        if self.choice != "op":
            # automatically add compositor connection
            _, _, op_t, op_add, _ = func_map["op"]
            find_op = self.find_by_type(nt, op_t)
            if not find_op:
                find_op = nt.nodes.new(type=op_add)
            nt.links.new(an_output, find_op.inputs[0])

        if self.choice == "ba":
            connections = [c.to_node for c in rn.outputs[0].links if c.to_node.type ==
                           "GROUP" and c.to_node.node_tree.name.startswith("Control")]
            if connections:
                nt.links.new(connections[0].outputs[0], find.inputs[1])
            else:
                nt.links.new(rn.outputs[0], find.inputs[1])
            nt.links.new(an_output, find.inputs[0])
        else:
            nt.links.new(an_output, find.inputs[0])

        # check existing links
        connected = list(set([
            l.to_node for l in an_output.links]))

        avg = sum([max(100, t.dimensions[1]) +
                  50 for t in connected]) / len(connected)
        offset = -avg
        for c in range(len(connected)):
            connected[c].location = (
                an.location.x + 250, an.location.y - offset)

            offset += max(connected[c].dimensions[1], 100) + 50
        nt.nodes.active = find
        find.select = True
        an.select = False

        if "VIEWER" in nt.nodes.active.type and context.space_data.type != "IMAGE_EDITOR":
            if not context.space_data.show_backdrop:
                context.space_data.show_backdrop = True

        return {"FINISHED"}


class ADD_OT_postfxinput(Operator):
    """Adds selected input to compositor"""
    bl_idname = "postfx.input"
    bl_label = "Add Input"

    choice: bpy.props.StringProperty()
    move_cursor: bpy.props.BoolProperty(default=True)

    def execute(self, context):

        non_append = ["rl", "im"]
        nt = context.space_data.edit_tree if context.space_data.type == "NODE_EDITOR" else context.scene.node_tree

        _, _, at, _ = func_map[self.choice]

        if self.move_cursor:
            cursor_loc = context.space_data.cursor_location
        else:
            cursor_loc = (0, -1000)

        new_n = None
        if self.choice not in non_append:
            new_n = append_nodegroup(context, self.choice)
        else:
            new_n = nt.nodes.new(type=at)
            new_n.use_custom_color = True
            new_n.color = color_lookup["red"]

        for sel in [n for n in nt.nodes if n.select]:
            sel.select = False

        if new_n:
            new_n.select = True
            nt.nodes.active = new_n
            new_n.location = (cursor_loc[0], cursor_loc[1])

            if self.choice == "im":
                new_n2 = append_nodegroup(context, "ic")
                new_n2.select = True
                nt.nodes.active = new_n2
                new_n2.location = (cursor_loc[0] + 200, cursor_loc[1])

                nt.links.new(new_n.outputs[0], new_n2.inputs[0])
            elif self.choice == "rl":
                new_n.width = 150
        if self.move_cursor:
            return bpy.ops.transform.translate('INVOKE_DEFAULT')
        else:
            return {"FINISHED"}


class ADD_OT_postfxnode(Operator):
    """Adds selected Shader to compositor"""
    bl_idname = "postfx.add"
    bl_label = "Add Node"

    choice: bpy.props.StringProperty()
    move_amount: bpy.props.IntProperty(default=200)

    @classmethod
    def poll(cls, context):
        return context.scene.node_tree and context.scene.node_tree.nodes.active and len([n for n in context.scene.node_tree.nodes.active.outputs if n.type == "RGBA"])

    def execute(self, context):
        nt = context.space_data.edit_tree if context.space_data.type == "NODE_EDITOR" else context.scene.node_tree
        pref = context.preferences.addons[__package__].preferences
        an = nt.nodes.active
        x, y = an.location
        from_socket = [
            o for o in an.outputs if o.type == "RGBA"][0]
        from_socket_links = [(l.to_node, l.to_socket.name)
                             for l in from_socket.links]

        connected_nodes = self.collect_connected_nodes(an)

        new_n = append_nodegroup(context, self.choice)

        # add input
        time = [i for i in new_n.inputs if i.name.lower() == "time"]
        for t in time:
            d = t.driver_add("default_value")
            d.driver.expression = "frame"

        for sel in [n for n in nt.nodes if n.select]:
            sel.select = False

        if new_n:
            new_n.select = True
            nt.nodes.active = new_n
            new_n.location = (x + 200, y)

            to_socket = [i for i in new_n.inputs if i.type == "RGBA"][0]

            # link previous node to new node
            nt.links.new(from_socket, to_socket)

            # link next node to new node
            if getattr(pref, "autoinsert"):
                for tn, tsn in from_socket_links:
                    nt.links.new(new_n.outputs[0], tn.inputs[tsn])

                # shift
                for cn in connected_nodes[:-1]:
                    cn.location.x += self.move_amount

        return {'FINISHED'}

    def collect_connected_nodes(self, current):
        if not len(current.outputs):
            return [current]
        else:
            result = []
            for o in current.outputs:
                for l in o.links:
                    result += self.collect_connected_nodes(l.to_node)
            result.append(current)
            return result


class COMPOSITOR_MT_postfx_input(Menu):
    # label is displayed at the center of the pie menu.
    bl_label = "Select Input"

    def check_input_exists(self, context, current_code):
        l, _, t, _ = func_map[current_code]
        sel = context.scene.node_tree.nodes.active
        sel_same = False
        if sel and sel.type == t:
            if t == "GROUP" and sel.node_tree.name == l:
                sel_same = True

        exists = False
        for n in context.scene.node_tree.nodes:
            if t != "GROUP" and n.type == t:
                exists = True

        return sel_same or exists

    def draw_io(self, iorow, context):
        box = iorow.box()
        baserow = box.row()
        col1 = baserow.column(align=True)
        row = col1.row()
        row.label(text="Input")
        input_nodes = ["rl", "im", "no", "co"]

        for ino in input_nodes:

            map_label, map_icon, _, _ = func_map[ino]
            if ino == "im":
                col1.separator()
                row = col1.column(align=True)
                row.scale_y = 1.25
                row.operator(ADD_OT_postfxinput.bl_idname,
                             text=map_label, icon=map_icon).choice = ino
                l, i, _, _ = func_map["ic"]
                row.operator(ADD_OT_postfxnode.bl_idname,
                             text=l, icon=i).choice = "ic"
                col1.separator()
            else:
                row = col1.row()
                row.scale_y = 1.25
                row.operator(ADD_OT_postfxinput.bl_idname,
                             text=map_label, icon=map_icon).choice = ino

        baserow.separator()
        # output
        col2 = baserow.column(align=True)
        row = col2.row()
        row.label(text="Output")
        output_nodes = ["op", "pr", "ba"]
        for on in output_nodes:
            l, i, _, _, _ = func_map[on]
            row = col2.row()
            row.scale_y = 1.25
            row.operator(ADD_OT_postfxoutput.bl_idname,
                         text=l, icon=i).choice = on
        col2.separator()
        row = col2.column(align=True)
        row.prop(context.space_data, "backdrop_zoom",
                 text="Zoom")
        row.operator("node.backimage_fit",
                     text="Fit Preview", icon="IMAGE_PLANE")

    def draw_adjust(self, adjustrow):
        box = adjustrow.box()
        baserow = box.row()
        cola1 = baserow.column(align=True)
        row = cola1.row()
        row.label(text="Detail")
        detail_nodes = ["br", "cl", "cn", "sp", "sa"]
        for dn in detail_nodes:
            map_label, map_icon, _ = func_map[dn]
            row = cola1.row()
            row.scale_y = 1.25
            row.operator(ADD_OT_postfxnode.bl_idname,
                         text=map_label, icon=map_icon).choice = dn
        cola2 = baserow.column(align=True)
        row = cola2.row()
        row.label(text="Color")

        color_nodes = ["ca", "se", "ti", "vi", "wa"]
        for cn in color_nodes:
            map_label, map_icon, _ = func_map[cn]
            row = cola2.row()
            row.scale_y = 1.25
            row.operator(ADD_OT_postfxnode.bl_idname,
                         text=map_label, icon=map_icon).choice = cn

        cola3 = baserow.column(align=True)
        row = cola3.row()
        row.label(text="Adjust")
        adjust_nodes = ["di", "fa", "gr", "le", "sh"]
        for an in adjust_nodes:
            map_label, map_icon, _ = func_map[an]
            row = cola3.row()
            row.scale_y = 1.25
            row.scale_x = 1.1
            row.operator(ADD_OT_postfxnode.bl_idname,
                         text=map_label, icon=map_icon).choice = an

    def draw_filters(self, filter_row):
        box = filter_row.box()
        col1 = box.column(align=True)
        col1.label(text="Filters")
        baserow = col1.row()
        for u in range(5):
            col = baserow.column(align=True)
            for v in range(2):
                ref = "f"+str(u+1 + v*5)
                row = col.row()

                map_label, map_icon, _ = func_map[ref]
                row.scale_y = 1.25
                row.operator(ADD_OT_postfxnode.bl_idname,
                             text=map_label, icon=map_icon).choice = ref

    def draw_effects(self, effects_row):
        box = effects_row.box()
        baserow = box.row()

        col1 = baserow.column(align=True)
        row = col1.row()
        row.label(text="Camera")

        cam_nodes = ["ln", "fr", "vg", "ts", "rb"]
        for cn in cam_nodes:
            map_label, map_icon, _ = func_map[cn]
            row = col1.row()
            row.scale_y = 1.25
            row.operator(ADD_OT_postfxnode.bl_idname,
                         text=map_label, icon=map_icon).choice = cn

        col3 = baserow.column(align=True)
        row = col3.row()
        row.label(text="Transform")
        t_nodes = ["sy", "tp"]
        for tn in t_nodes:
            map_label, map_icon, _ = func_map[tn]
            row = col3.row()
            row.scale_y = 1.25
            row.operator(ADD_OT_postfxnode.bl_idname,
                         text=map_label, icon=map_icon).choice = tn
        row = col3.row()
        row.label(text="Glow")
        g_nodes = ["fl", "bl"]
        for gn in g_nodes:
            map_label, map_icon, _ = func_map[gn]
            row = col3.row()
            row.scale_y = 1.25
            row.operator(ADD_OT_postfxnode.bl_idname,
                         text=map_label, icon=map_icon).choice = gn
        col2 = baserow.column(align=True)
        row = col2.row()
        row.label(text="Modify")
        m_nodes = ["ar","gl", "px", "ps", "de"]
        for mn in m_nodes:
            map_label, map_icon, _ = func_map[mn]
            row = col2.row()
            row.scale_y = 1.25
            row.operator(ADD_OT_postfxnode.bl_idname,
                         text=map_label, icon=map_icon).choice = mn

    def draw(self, context):
        layout = self.layout

        pie = layout.menu_pie()
        # input output
        iorow = pie.row()
        self.draw_io(iorow, context)

        effectsrow = pie.row()
        self.draw_effects(effectsrow)

        adjustrow = pie.row()
        self.draw_adjust(adjustrow)

        filterrow = pie.row()
        self.draw_filters(filterrow)


class POSTFX_OT_callpie(Operator):
    """Invoke the pie menu from shortcut"""
    bl_idname = "postfx.callpie"
    bl_label = "Calls pie menu"

    @classmethod
    def poll(self, context):
        return context.space_data.tree_type == "CompositorNodeTree"

    def execute(self, context):

        pref = context.preferences.addons[__package__].preferences
        scn = context.scene
        if not scn.use_nodes:
            scn.use_nodes = True

        nt = scn.node_tree
        if getattr(pref, "optimize"):
            optimize_setting(context)
            viewer_nodes = [
                n for n in nt.nodes if "VIEWER" in n.type]
            if not viewer_nodes:
                context.space_data.show_backdrop = False

            rlayers = [n for n in nt.nodes if n.type ==
                       "R_LAYERS" and not n.use_custom_color]
            for r in rlayers:
                r.width = 150
        bpy.ops.wm.call_menu_pie(name="COMPOSITOR_MT_postfx_input")

        return {'FINISHED'}


class POSTFX_OT_saveimage(Operator):
    """Save Image"""
    bl_idname = "postfx.saveimage"
    bl_label = "Save Output"

    @classmethod
    def poll(cls, context):
        global status
        return "Render Result" in bpy.data.images and bpy.data.images["Render Result"].has_data and not status

    def execute(self, context):
        scn = context.scene
        scn.render.image_settings.file_format = 'PNG'
        scn.render.image_settings.color_mode = 'RGBA'
        scn.render.image_settings.compression = 100

        bpy.ops.render.render(animation=False, write_still=False,
                              use_viewport=False, layer="", scene="")

        folder = bpy.path.abspath(scn.render.filepath)
        save_name = ""

        if os.path.isdir(folder):
            files = [f for f in os.listdir(
                folder) if os.path.isfile(os.path.join(folder, f))]

            count = 0
            while count < 200:
                save_name = "POSTFX_Output" + str(count + 1) + ".png"
                if save_name not in files:
                    break
                count += 1
        else:
            head, tail = os.path.split(folder)
            folder = head
            save_name = tail
            if not tail.endswith(".png"):
                save_name += ".png"

        path = os.path.join(folder, save_name)
        bpy.data.images["Render Result"].save_render(
            filepath=path)
        self.report({'INFO'}, 'Saved Image to: '+path)
        return {"FINISHED"}


class POSTFX_OT_modifylayer(Operator):
    bl_idname = "postfx.modifylayer"
    bl_label = "Add / Remove Layers"

    add: bpy.props.BoolProperty(default=True)

    @classmethod
    def poll(cls, context):
        global is_rendering
        return not is_rendering

    def execute(self, context):
        global subscribed
        scn = context.scene
        nt = scn.node_tree
        pref = context.preferences.addons[__package__].preferences
        if not scn.use_nodes:
            scn.use_nodes = True
            nt = context.scene.node_tree
            if len(nt.nodes) == 2:
                # remove default
                for n in nt.nodes:
                    nt.nodes.remove(n)
        if getattr(pref, "optimize"):
            optimize_setting(context)
        layer_list = scn.postfx_layers
        layer_index = scn.postfx_active_layer
        # add
        if self.add:
            if len(layer_list) == 0:
                setup_subscriptions(None, context)
            context.space_data.image = bpy.data.images["Render Result"]
            added = layer_list.add()
            added.id = scn.postfx_layer_id
            scn.postfx_layer_id += 1
            added.name = "None"
            scn.postfx_active_layer = len(layer_list) - 1
        # remove
        elif layer_index > -1:
            active_node = nt.nodes[layer_list[layer_index]
                                   .node_name] if layer_list[layer_index].node_name in nt.nodes else None
            if active_node:
                if active_node.outputs[0].links and scn.postfx_active_layer > 0:
                    prev_sock = active_node.inputs[0].links[0].from_socket
                    next_socks = [
                        l.to_socket for l in active_node.outputs[0].links]
                    for s in next_socks:
                        nt.links.new(
                            prev_sock.node.outputs[prev_sock.name], s.node.inputs[s.name])
                if scn.postfx_active_layer > 0:
                    nt.nodes.remove(
                        nt.nodes[layer_list[layer_index].node_name])
                else:
                    if layer_index == 0:
                        # remove comp nodes
                        for l in active_node.outputs[0].links:
                            nt.nodes.remove(l.to_node)
                        if active_node.type == "GROUP":
                            nt.nodes.remove(
                                active_node.inputs[0].links[0].from_node)
                    nt.nodes.remove(active_node)
            layer_list.remove(layer_index)
            new_idx = min(max(0, layer_index - 1), len(layer_list) - 1)
            scn.postfx_active_layer = new_idx
            if (getattr(pref, "autorefresh")):
                return bpy.ops.postfx.refresh()
        return {"FINISHED"}


class POSTFX_OT_shiftlayer(Operator):
    bl_idname = "postfx.shiftlayer"
    bl_label = "Shift Layers"

    direction: bpy.props.IntProperty(default=1)

    @classmethod
    def poll(cls, context):
        global is_rendering
        return not is_rendering

    def collect_io(self, io, key):
        coll = []
        for i in [i for i in io if len(i.links)]:
            for l in i.links:
                coll.append((i.name, getattr(l, key + "_node"),
                            getattr(l, key + "_socket").name))
        return coll

    def apply_connections(self, nt, n, io, is_input):
        src_key = "inputs" if is_input else "outputs"
        dest_key = "outputs" if is_input else "inputs"

        for entry in io:
            src_sock, dest_node, dest_sock = entry
            if dest_node.name != n.name:
                nt.links.new(getattr(n, src_key)[src_sock], getattr(
                    dest_node, dest_key)[dest_sock])

    def execute(self, context):
        scn = context.scene
        nt = scn.node_tree
        layers = scn.postfx_layers
        pref = context.preferences.addons[__package__].preferences
        idx = scn.postfx_active_layer

        new_idx = scn.postfx_active_layer + self.direction
        old_node = layers[idx].node_name
        swap_node = layers[new_idx].node_name

        layers.move(new_idx, idx)

        # set new index
        scn.postfx_active_layer = min(max(0, new_idx), len(layers) - 1)

        on = nt.nodes[old_node]
        sn = nt.nodes[swap_node]
        temp_x, temp_y = sn.location
        sn.location = on.location
        on.location = (temp_x, temp_y)

        # inputs
        old_inputs = self.collect_io(on.inputs, "from")
        swap_inputs = self.collect_io(sn.inputs, "from")

        old_outputs = self.collect_io(on.outputs, "to")
        swap_outputs = self.collect_io(sn.outputs, "to")

        if idx < new_idx:
            nt.links.new(sn.outputs[0], on.inputs[0])
        else:
            nt.links.new(on.outputs[0], sn.inputs[0])

        # input
        self.apply_connections(nt, on, swap_inputs, True)
        self.apply_connections(nt, sn, old_inputs, True)

        # output
        self.apply_connections(nt, on, swap_outputs, False)
        self.apply_connections(nt, sn, old_outputs, False)
        if (getattr(pref, "autorefresh")):
            return bpy.ops.postfx.refresh()
        return {"FINISHED"}


class POSTFX_layerprops(bpy.types.PropertyGroup):
    name: bpy.props.StringProperty(
        name="", default="NAME OF SLOT", description="Name")
    node_name: bpy.props.StringProperty(
        name="", description="Name")
    icon: bpy.props.StringProperty(
        default="ERROR", description="Icon")
    layer_category: bpy.props.EnumProperty(
        name="Category", items=get_layer_categories, update=assign_new_category)
    layer_type: bpy.props.EnumProperty(
        name="Type", items=get_layer_types, update=assign_new_layer)
    id: bpy.props.IntProperty(default=0)
    type: bpy.props.StringProperty(
        name="", default="n", description="Type")
    visible: bpy.props.BoolProperty(
        name="Visible", default=True, description="Visibility", update=change_visibility)


class POSTFX_UL_layers(UIList):

    def draw_item(self, context, layout, data, item, icon, active_data, active_propname, index):
        # draw_item must handle the three layout types... Usually 'DEFAULT' and 'COMPACT' can share the same code.
        if self.layout_type in {'DEFAULT', 'COMPACT'}:
            if item:
                row = layout.row()
                # row.scale_y = 1.5
                row.prop(item, "name", text="",
                         emboss=False, icon=getattr(item, "icon"))
                row.use_property_decorate = False
                if index != 0 and item.layer_type != "n":
                    if item.visible:
                        row.prop(item, "visible", text="",
                                 emboss=False, icon='HIDE_OFF')
                    else:
                        row.prop(item, "visible", text="",
                                 emboss=False, icon='HIDE_ON')
            else:
                layout.label(text="", translate=False)
        # 'GRID' layout type should be as compact as possible (typically a single icon!).
        elif self.layout_type in {'GRID'}:
            layout.alignment = 'CENTER'

    def filter_items(self, context, data, propname):
        items = getattr(data, propname).items()
        enum = list(enumerate(items[::-1]))
        sort_items = bpy.types.UI_UL_list.sort_items_helper
        ordered = sort_items(enum, lambda i: i[0], True)
        return [], ordered


class POSTFX_OT_refresh(Operator):
    bl_idname = "postfx.refresh"
    bl_label = "Refresh Output"

    _timer = None
    _first_pass = False

    def modal(self, context, event):
        if (context.area):
            context.area.tag_redraw()
        if event.type in {'ESC'}:
            self.cancel(context)
            return {'CANCELLED'}

        if event.type == 'TIMER':

            global is_rendering
            if not is_rendering and self._first_pass:
                self._first_pass = False
                bpy.ops.render.render(animation=False, write_still=False,
                                      use_viewport=False, layer="", scene="")
            elif not is_rendering and not self._first_pass:
                self.cancel(context)
                return {"FINISHED"}
        return {'PASS_THROUGH'}

    def cancel(self, context):
        wm = context.window_manager
        self.reset_props(context)
        wm.event_timer_remove(self._timer)

    def execute(self, context):
        wm = context.window_manager
        self._timer = wm.event_timer_add(
            TIME_INTERVAL, window=context.window)
        wm.modal_handler_add(self)
        self._first_pass = True
        global status
        status = "Compositing..."
        return {'RUNNING_MODAL'}

    def reset_props(self, context):
        self._first_pass = False
        global status
        status = ""


class POSTFX_PT_layerpanel(Panel):
    bl_space_type = 'IMAGE_EDITOR'
    bl_region_type = 'UI'
    bl_category = 'Post FX'
    bl_label = 'Post FX'

    def draw(self, context):
        global status
        scn = context.scene
        nt = scn.node_tree
        layers = scn.postfx_layers
        pref = context.preferences.addons[__package__].preferences
        active_idx = scn.postfx_active_layer
        layout = self.layout
        row = layout.row()
        row.enabled = not status
        row.template_list("POSTFX_UL_layers", "", scn,
                          "postfx_layers", scn, "postfx_active_layer")

        if len(layers) > 1 and active_idx != 0 and layers[0].type == "n":
            row = layout.row()
            row.label(text="Input Layer Unassigned", icon="ERROR")
        else:
            sidebar = row.column(align=True)
            sidebar.enabled = not status
            # add remove
            sidebar.operator(POSTFX_OT_modifylayer.bl_idname,
                             icon="ADD", text="").add = True
            if not (active_idx == 0 and len(layers) > 1):
                sidebar.operator(POSTFX_OT_modifylayer.bl_idname,
                                 icon="REMOVE", text="").add = False
            sidebar.separator()
            # shift
            sidebar.separator()
            if active_idx > 0 and active_idx < len(layers) - 1:
                sidebar.operator(POSTFX_OT_shiftlayer.bl_idname,
                                 icon="TRIA_UP", text="").direction = 1
            if active_idx > 1:
                sidebar.operator(POSTFX_OT_shiftlayer.bl_idname,
                                 icon="TRIA_DOWN", text="").direction = -1
            layout.separator()
            if (layers):
                active = layers[active_idx]
                if active_idx > 0:
                    row = layout.row()
                    row.enabled = not status
                    split = row.split(factor=0.35)
                    split.label(text="Category")
                    split.prop(active, "layer_category", text="")
                row = layout.row()
                row.enabled = not status
                if active_idx > 0:
                    split = row.split(factor=0.35)
                    split.label(text="Type")
                    split.prop(active, "layer_type", text="")
                else:
                    row.prop(active, "layer_type", text="Input")
                layout.separator()

                if active.node_name and active.node_name in nt.nodes:
                    box = layout.box()
                    row = box.row()
                    row.label(text="Properties")
                    active_node = nt.nodes[active.node_name]
                    if active_idx > 0:
                        for i in [i for i in active_node.inputs if len(i.links) == 0]:
                            row = box.row()
                            row.prop(i, "default_value", text=i.name)
                    else:
                        if active.type in prop_map:
                            for p in prop_map[active.type]:
                                row = box.row()
                                if p not in active_node and p == "image":
                                    row.template_ID(
                                        active_node.inputs[0].links[0].from_node, p, open="image.open")
                                else:
                                    row.prop(active_node, p)
                        col = box.column()
                        for i in [i for i in active_node.inputs if len(i.links) == 0]:
                            col.prop(i, "default_value", text=i.name)
                    row = box.row()
                    if not status:
                        if getattr(pref, "autorefresh"):
                            row.prop(pref, 'autorefresh')
                        else:
                            row.prop(pref, 'autorefresh')
                            row = box.row()
                            row.scale_y = 1.5
                            row.operator("postfx.refresh", text="Refresh",
                                     icon="FILE_REFRESH")
                    else:
                        row.scale_y = 1.5
                        row.label(text=status, icon="INFO")
        if layers and layers[0].type != "n":
            layout.separator()
            box = layout.box()
            box.enabled = not status
            row = box.row()
            row.prop(scn.render, "filepath", text="")
            row = box.row()
            row.operator("render.render", text="Re-Render",
                         icon="RENDER_STILL")
            row = box.row()
            row.scale_y = 1.25
            row.operator("postfx.saveimage", text="Save Image",
                         icon="EXPORT")


class POSTFX_PT_optionspanel(Panel):
    bl_space_type = 'NODE_EDITOR'
    bl_region_type = 'UI'
    bl_category = 'Post FX'
    bl_label = 'Post FX'

    @classmethod
    def poll(cls, context):
        return context.space_data.tree_type == "CompositorNodeTree"

    def draw(self, context):
        pref = context.preferences.addons[__package__].preferences
        layout = self.layout
        row = layout.row()
        if context.scene.render.engine not in ["CYCLES", "BLENDER_EEVEE"]:
            row.label(text="Only Cycles / Eevee supported")
        else:
            if context.space_data.tree_type == "CompositorNodeTree":
                if not context.scene.node_tree:
                    row.label(text="Enable Nodes")
                else:
                    row.prop(pref, 'optimize')
                    row = layout.row()
                    row.prop(pref, 'autoinsert')
                    row = layout.row()
                    row.operator(POSTFX_OT_callpie.bl_idname,
                                 text="Invoke Pie Menu", icon="TRIA_RIGHT")

def setup_subscriptions(self, context):
    global subscribed
    global sub_owner
    pref = context.preferences.addons[__package__].preferences
    if (getattr(pref, "autorefresh") and not subscribed):
        #add sub
        bpy.msgbus.subscribe_rna(
            key=(bpy.types.NodeSocketFloat, "default_value"),
            owner=sub_owner,
            args=(context,),
            notify=msgbus_update,
        )
        bpy.msgbus.subscribe_rna(
            key=(bpy.types.NodeSocketColor, "default_value"),
            owner=sub_owner,
            args=(context,),
            notify=msgbus_update,
        )
        bpy.msgbus.subscribe_rna(
            key=(bpy.types.NodeSocketFloatFactor, "default_value"),
            owner=sub_owner,
            args=(context,),
            notify=msgbus_update,
        )
        subscribed = True
    else:
        #remove sub
        bpy.msgbus.clear_by_owner(sub_owner)
        subscribed = False

class POSTFX_PR_preferences(bpy.types.AddonPreferences):
    bl_idname = __package__

    optimize: bpy.props.BoolProperty(
        default=True, name="Auto Optimize Preview", description="Sets optimal settings for performance and viewport")
    autoinsert: bpy.props.BoolProperty(
        default=True, name="Auto Insertion", description="Inserts new nodes between the active node and connected output")
    autorefresh: bpy.props.BoolProperty(default=True, name="Auto Refresh", description="Whether to automatically refresh image editor updates", update=setup_subscriptions)

    def draw(self, context):
        layout = self.layout
        col = layout.column(align=True)
        row = col.row()
        row.scale_y = 1.25
        row.label(text="Invoke Post FX Shortcut")
        km, kmi = addon_keymaps[0]
        row.split(factor=0.25, align=True)
        row.prop(kmi, "value", text="")
        row.prop(kmi, "type", text="", full_event=True)

        col = layout.column()
        row = col.row()
        row.prop(self, "optimize")
        row.prop(self, "autoinsert")
        row.prop(self, "autorefresh")
        row = layout.separator()


addon_keymaps = []
classes = (
    COMPOSITOR_MT_postfx_input,
    POSTFX_OT_callpie,
    ADD_OT_postfxinput,
    ADD_OT_postfxnode,
    ADD_OT_postfxoutput,
    POSTFX_PT_layerpanel,
    POSTFX_PT_optionspanel,
    POSTFX_PR_preferences,
    POSTFX_UL_layers,
    POSTFX_layerprops,
    POSTFX_OT_modifylayer,
    POSTFX_OT_shiftlayer,
    POSTFX_OT_saveimage,
    POSTFX_OT_refresh
)


@persistent
def set_render_state_true(scene):
    global is_rendering
    is_rendering = True


@persistent
def set_render_state_false(scene):
    global is_rendering
    is_rendering = False


def register():
    from bpy.utils import register_class
    for cls in classes:
        try:
            register_class(cls)
        except:
            print("[PostFX]: Problem registering class, potential conflict with free version")

    # Add the hotkey
    wm = bpy.context.window_manager
    kc = wm.keyconfigs.addon
    if kc:
        km = wm.keyconfigs.addon.keymaps.new(
            name='Node Editor', space_type='NODE_EDITOR')
        kmi = km.keymap_items.new(
            POSTFX_OT_callpie.bl_idname, type='A', value='PRESS', ctrl=True)
        addon_keymaps.append((km, kmi))

    bpy.types.Scene.postfx_layers = bpy.props.CollectionProperty(
        type=POSTFX_layerprops)
    bpy.types.Scene.postfx_active_layer = bpy.props.IntProperty(
        name="Active Layer", default=0, update=layer_selection)
    bpy.types.Scene.postfx_layer_id = bpy.props.IntProperty(
        name="Active Layer ID", default=0)
    bpy.app.handlers.render_init.append(set_render_state_true)
    bpy.app.handlers.render_cancel.append(set_render_state_false)
    bpy.app.handlers.render_complete.append(set_render_state_false)


def unregister():
    from bpy.utils import unregister_class
    for cls in classes:
        try:
            unregister_class(cls)
        except:
            print("[PostFX]: Problem unregistering class, potential conflict with free version")
    
    global sub_owner
    bpy.msgbus.clear_by_owner(sub_owner)

    # Remove the hotkey
    for km, kmi in addon_keymaps:
        km.keymap_items.remove(kmi)
    addon_keymaps.clear()
    if ( getattr(bpy.types.Scene, "postfx_layers", None)):
        del bpy.types.Scene.postfx_layers
    if (getattr(bpy.types.Scene, "postfx_active_layer", None)):
        del bpy.types.Scene.postfx_active_layer
    if (getattr(bpy.types.Scene, "postfx_layer_id", None)):
        del bpy.types.Scene.postfx_layer_id
    bpy.app.handlers.render_init.remove(set_render_state_true)
    bpy.app.handlers.render_cancel.remove(set_render_state_false)
    bpy.app.handlers.render_complete.remove(set_render_state_false)
